/*-
 * Copyright (c) 2008 Vincenzo Iozzo <snagg@sikurezza.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * To run the code:
 * gcc -o builder build.c loader.c
 * gcc -o jumper jumper.c
 * ./builder <hostname> <port> <x86Macho> "<argv1>:<argv2>..:" "<envp1>:<envp2>..:" [<exec_string>]
 * ./jumper <port>
 */

#include <mach-o/loader.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/uio.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <netinet/in.h>
#include <arpa/inet.h>

extern void loader();
extern void loader_end();
extern void begin();

#define ENTRY_POINT (loader - begin)
#define LOADER_CODESZ (loader_end - begin)
#define	PG_SIZE		0x1000
#define	TOP_STACK	0xC0000000


char **split(char *str, char ch) {
	char **ret;
	char *start, *end;
	int i;
	int num;
	
	num = 1;
	for (i = 0 ; str[i] ; i++)
		if (str[i] == ch) num++;
	
	ret = (char **)malloc(sizeof(char *) * num + 1);
	
	start = str;
	for (i = 0 ; i < num ; i++) {
		end = strchr(start, ch);
		if (end) *end = 0;
		ret[i] = start;
		start = end + 1;
	}
	ret[num] = NULL;
	
	return(ret);
}

char *build_stack(char *stack, char *path, char **argv, char **envp, unsigned long *stack_len) {
	char *ptr;
	int argc, envc;
	int i, n;
	unsigned long macho_offset;
	char *path_addr;
	extern char **environ;
	char **environ_val;
	
	macho_offset = 0x1000;
	
	memset(stack, 0, PG_SIZE);
	ptr = stack + PG_SIZE;
	
	ptr -= sizeof(unsigned long);
	*ptr = 0; 

	for (argc = 0 ; argv[argc] ; argc++);
	for (envc = 0 ; envp[envc] ; envc++);
	for(n = 0; environ[n]; n++);
	
	environ_val = malloc(n * sizeof(unsigned long));
	
	for (i = envc - 1 ; i >= 0 ; i--) {
		ptr -= strlen(envp[i]) + 1;
		strcpy(ptr, envp[i]);
		envp[i] = (char *)(TOP_STACK - (stack + PG_SIZE - ptr));
		// printf("envp[%i] = %p\n", i, envp[i]);
	}
	
	for (i = n - 1 ; i >= 0 ; i--) {
		ptr -= strlen(environ[i]) + 1;
		strcpy(ptr, environ[i]);
		environ_val[i] = (char *)(TOP_STACK - (stack + PG_SIZE - ptr));
	}
	
	for (i = argc - 1 ; i >= 0 ; i--) {
		ptr -= strlen(argv[i]) + 1;
		strcpy(ptr, argv[i]);
		argv[i] = (char *)(TOP_STACK - (stack + PG_SIZE - ptr));
		// printf("argv[%i] = %p\n", i, argv[i]);
	}
		
	ptr -= strlen(path) +1;
	strcpy(ptr, path);
	path_addr = (char *)(TOP_STACK - (stack + PG_SIZE -ptr));
	
	ptr = (char *)((unsigned long)ptr & 0xfffffff0);

	ptr -= sizeof(unsigned long) * 8;

	ptr -= sizeof(unsigned long);
	memcpy(ptr, &path_addr, sizeof(unsigned long));
	
	ptr -= (envc + 1) * sizeof(unsigned long);
	memcpy(ptr, envp, (envc + 1) * sizeof(unsigned long));
	
	ptr -= n  * sizeof(unsigned long);
	memcpy(ptr, environ_val, n  * sizeof(unsigned long));
	
	ptr -= (argc + 1) * sizeof(unsigned long);
	memcpy(ptr, argv, (argc + 1) * sizeof(unsigned long));
	
	ptr -= sizeof(unsigned long);
	*ptr = argc;
	ptr -= sizeof(unsigned long);
	memcpy(ptr, &macho_offset, sizeof(unsigned long));
	
	*stack_len = TOP_STACK - (stack + PG_SIZE- ptr);

	return ptr;
}


void *map_file(char *filename, int *sz) 
{
	struct stat st;
	void *ret;
	int fd;
	
	if ((fd = open(filename, O_RDONLY)) == -1)
		return(NULL);
	
	if (fstat(fd, &st) == -1) {
		close(fd);
		return(NULL);
	}
	
	*sz = st.st_size;
	
	ret = mmap(NULL, *sz, PROT_READ, MAP_PRIVATE, fd, 0);
	
	return ret;
}

int main(int argc, char *argv[])
{
	void *orig, *new; 
	struct mach_header *header;
	int offset, original_size, delta, new_size;
	int i,k, align, div, sock;
	char **arg, **env;
	char *stack_data;
	unsigned long stack;
	struct load_command *cmd;
	struct segment_command *seg; 
	struct section *sec, *orig_sec; 
	struct sockaddr_in srv;
	
	new_size = offset = 0;
	
	if(!(orig = map_file(argv[3], &original_size))) {
		printf("Failed to map the file\n");
		exit(-1);
	}
	
	header = malloc(sizeof(struct mach_header));
	
	memcpy(header, orig, sizeof(struct mach_header));
	
	// Intel only arch, sorry
	if (((header->cputype & ~CPU_ARCH_MASK) != 0x7)) {
		printf("Wrong arch\n");
		exit(-1);
	}
	
	if(header->filetype != MH_EXECUTE) {
		printf("File not supported\n");
		exit(-1);
	}
	
	stack_data = malloc(PG_SIZE);
	if(stack_data == NULL) {
		printf("Unable to allocate the new stack\n");
		exit(-1);
	}
	
	arg = split(argv[4], ':');
	env = split(argv[5], ':');
	if(argc == 7)
		stack_data = build_stack(stack_data, argv[6], arg, env, &stack);
	else
		stack_data = build_stack(stack_data, argv[3], arg, env, &stack);
		
	header->sizeofcmds += sizeof(struct section);
	
	div = original_size % PG_SIZE;
	align = PG_SIZE - div; 
	
	new_size = original_size + LOADER_CODESZ + (TOP_STACK - stack) + align;
	new = malloc(new_size);
	if (new == NULL) {
		printf("Failed to map the new file\n");
		exit(-1);
	}
	memset(new, 0, new_size);	
	memcpy(new, header, sizeof(struct mach_header));
	
	offset = sizeof(struct mach_header);
	
	for(i = 0; i< header->ncmds; i++) {
		cmd = orig + offset;
		if(cmd->cmd == 0x1) {
			seg = malloc(sizeof(struct segment_command));
			memcpy(seg, orig + offset, sizeof(struct segment_command));
			
			if(!strncmp(seg->segname, "__PAGEZERO", strlen(seg->segname))) {
				
				/*
				 * Change offset and protection. Add a section
				 */
				seg->cmdsize += sizeof(struct section);
				seg->fileoff = original_size + align;
				seg->filesize = LOADER_CODESZ + (TOP_STACK - stack);
				seg->maxprot = 0x31337;
				//seg->initprot = 0x5; 
				seg->nsects = 0x1;
				memcpy(new + offset, seg, sizeof(struct segment_command));
				offset +=sizeof(struct segment_command);
				
				/*
				 * Create the new section.
				 */
				sec = malloc(sizeof(struct section));
				if(sec == NULL) {
					printf("Unable to allocate the new section\n");
					exit(-1);
				}	
				strcpy(sec->sectname, "crafted");
				strcpy(sec->segname, "crafted");
				sec->offset = LOADER_CODESZ;
				sec->addr = stack;
				sec->size = TOP_STACK - stack;
				sec->align = 0x2;
				sec->flags = 0x80000000;
				memcpy(new + offset, sec, sizeof(struct section));
				break;
			}
			memcpy(new + offset, seg, sizeof(struct segment_command));
			offset +=sizeof(struct segment_command);
			
			for(k = 0; k <seg->nsects; k++) {
				orig_sec = orig + offset;
				memcpy(new + offset, orig_sec, sizeof(struct section));
				offset += sizeof(struct section);
			}
		}
		memcpy(new + offset, orig + offset, cmd->cmdsize);
		offset += cmd->cmdsize;
	}
	
	//Write the address of the loader stub
	header->magic = original_size + align + ENTRY_POINT;
	memcpy(new, header, sizeof(struct mach_header));
	
	/*
	 * the layout of the new file at the moment is:
	 * MACH HEADER
	 * SEGMENT0
	 * ...
	 * SEGMENTN
	 * PAGEZERO SEGMENT
	 * PAGEZERO SECTION
	 *
	 * Now we need to copy the rest of the file.
	 * We delete sizeof(struct section) bytes just after the commands, and then copy the rest of the file.
	 *
	 */
	delta = header->sizeofcmds - offset - sizeof(struct section) + sizeof(struct mach_header);
	memcpy(new + offset + sizeof(struct section), orig + offset, delta);
	offset += delta;
	
	//copy the rest
	memcpy(new + offset + sizeof(struct section) , orig + offset +  sizeof(struct section), original_size - offset - sizeof(struct section));
	
	
	//Copy the loader at the bottom of the file
	memcpy(new + original_size  + align, begin, LOADER_CODESZ);
	memcpy(new + original_size + sec->offset + align , stack_data, TOP_STACK - stack);
	
	
	srv.sin_family = AF_INET;
	srv.sin_port = htons(atoi(argv[2]));
	srv.sin_addr.s_addr = inet_addr(argv[1]);
	
	if ((sock = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		perror("socket()");
		exit(errno);
	}
	
	if (connect(sock, (struct sockaddr *)&srv, sizeof(struct sockaddr)) == -1) {
		perror("connect()");
		exit(errno);
	}
	
	/* send object to jumper */	
	write(sock, new, new_size);
	
	close(sock);
	free(new);
	free(sec);
	
	return 0;
}