/*-
 * Copyright (c) 2008 Vincenzo Iozzo <snagg@sikurezza.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <mach-o/loader.h>
#include <mach-o/nlist.h>
#include <mach-o/fat.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/uio.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <errno.h>

#include <sys/syscall.h>

void begin()
{
}

void *map_library() 
{
	struct stat st;
	void *ret;
	int fd;
	int err;
	
	//fd = open("/usr/lib/libSystem.B.dylib", O_RDONLY);
	__asm__ __volatile__ (
		"subl	$52, %%esp\n"
		"pushl	$0x00006269\n"
		"pushl	$0x6c79642e\n"
		"pushl	$0x422e6d65\n"
		"pushl	$0x74737953\n"
		"pushl	$0x62696c2f\n"
		"pushl	$0x62696c2f\n"
		"pushl	$0x7273752f\n"
		"movl	%%esp, %%edx\n"
		"pushl	$0x0\n"
		"pushl	%%edx\n"
		"xorl %%eax, %%eax\n"
		"movb $5, %%al\n"
		"pushl %%eax\n"
		"int $0x80\n"
		"movl %%eax, %0"
		:"=m"(fd) 
		:
		);

	if(fd == -1)	
		return(NULL);
	
	//err = fstat(fd, &st);
	__asm__ __volatile__ (
		"leal %2, %%eax\n"
		"movl %%eax, 4(%%esp)\n"
		"movl %1, %%eax\n"
		"movl %%eax, (%%esp)\n" 
		"xorl %%eax, %%eax\n"
		"movb $189, %%al\n"
		"pushl %%eax\n"
		"int $0x80\n"
		"movl %%eax, %0"
		:"=m"(err) 
		:"m"(fd), "m"(st)
		);
	
	if (err == -1)
		return(NULL);
	
	//ret = mmap(NULL, st.st_size, PROT_READ, MAP_PRIVATE, fd, 0);
	__asm__ __volatile__ (
		"movl	$0, 24(%%esp)\n"
		"movl	$0, 20(%%esp)\n"
		"movl	%2, %%eax\n"
		"movl	%%eax, 16(%%esp)\n" 
		"movl	$2, 12(%%esp)\n"
		"movl	$1, 8(%%esp)\n"
		"movl	%1, %%eax\n"
		"movl	%%eax, 4(%%esp)\n"
		"movl	$0, (%%esp)\n"
		"xorl	%%eax, %%eax\n"
		"movb	$197, %%al\n"
		"pushl	%%eax\n"
		"int	$0x80\n"
		"mov	%%eax, %0\n"
		:"=m"(ret) 
		:"m"(st.st_size), "m"(fd)
		);
	
	return ret;
}

unsigned int calculate_djb2(char *string)
{
	unsigned int h = 5381;
	int c;
	
	while(c = *string++) 
		h = ((h << 5) + h )+ c;
	
	return h;
}

int *retrive_symbols(void *binary, void *virtual_binary, unsigned int *symbols, int number)
{
	struct mach_header *header; 
	struct load_command *cmd; 
	struct nlist *symbol; 
	struct symtab_command *sym;
	struct segment_command *seg;
	struct fat_header *fat; 
	struct fat_arch *arch;
	char *sym_name; 
	long base_data;
	int offset, i, k, found, string_off, base; 
	int ret[6];
	unsigned int data_hash = 0xec5699bd;
	unsigned int linkedit_hash = 0xc214bfb7;
	unsigned int hash;
		
	offset = found = 0; 
	fat = binary;
	
	offset +=sizeof(struct fat_header);
	
	for(i = 0; i < fat->nfat_arch; i++) {
		arch = binary + offset; 
		if(arch->cputype == 0x7000000)
			break;
		offset += sizeof(struct fat_arch);
	}	
	
	offset = 0;
	base = (arch->offset>>24) | 
        ((arch->offset<<8) & 0x00FF0000) |
        ((arch->offset>>8) & 0x0000FF00) |
        (arch->offset<<24);
	offset = base;
	header = binary + offset; 
 
    offset += sizeof(struct mach_header);
	
	for(i = 0; i< header->ncmds; i++) {
		cmd = binary + offset; 
		if(cmd->cmd == 0x2) {
			sym = binary + offset; 
			if(found == 2)
				break;
				
		}else if(cmd->cmd == 0x1) {
			if(found == 2) {
				offset += cmd->cmdsize;
				continue;
			}
			seg = binary + offset;
			
			hash = calculate_djb2(seg->segname);
			
			if(hash == data_hash){
				base_data = seg->vmaddr;
				found++;	
			}else if(hash == linkedit_hash)
				found++;
		}		
		offset += cmd->cmdsize;

	}
	
	offset = base + sym->symoff;
	string_off = base + sym->stroff;
	
	for(i = 0; i < sym->nsyms; i++) {
		symbol = binary + offset;
		offset += sizeof(struct nlist);
		sym_name = (char *)binary + symbol->n_un.n_strx + string_off;
		if(symbol->n_un.n_strx == 0x0)
			continue;
		hash = calculate_djb2(sym_name);
		for( k = 0 ; k < number ; k++) {
			if(hash == symbols[k])
				ret[k] = symbol->n_value - base_data;	
		}
	}
	header = virtual_binary;
	offset = 0;
	
	offset +=sizeof(struct mach_header);
	
	for(i = 0; i< header->ncmds; i++) {
		cmd = virtual_binary + offset; 
		if(cmd->cmd == 0x1) {
			seg = virtual_binary + offset;
			if(calculate_djb2(seg->segname) == data_hash)
				break;
		}		
		offset += cmd->cmdsize;
	}
	
	for(k = 0; k < number; k++)
		ret[k] = ret[k] + seg->vmaddr;

	return ret;
}

	
int *find_symbols(void *binary, unsigned int *symbols, int number)
{
	struct mach_header *header; 
	struct load_command *cmd; 
	struct nlist *symbol; 
	struct symtab_command *sym;
	struct segment_command *seg;
	char *sym_name; 
	int offset, i, k, found, string_off; 
	int ret[8];
	int *ret_val;
	unsigned int linkedit_hash = 0xc214bfb7;
	unsigned int hash;
		
	offset = found = 0; 
	header = binary; 
	offset += sizeof(struct mach_header);
	
	for(i = 0; i< header->ncmds; i++) {
		cmd = binary + offset; 
		if(cmd->cmd == 0x2) {
			sym = binary + offset; 
			if(found)
				break;
				
		}else if(cmd->cmd == 0x1) {
			if(found) {
				offset += cmd->cmdsize;
				continue;
			}
			seg = binary + offset;
			
			if(calculate_djb2(seg->segname) == linkedit_hash)
				found = 1;
		}		
		offset += cmd->cmdsize;
	}
	
	offset = sym->symoff - seg->fileoff;
	string_off = sym->stroff - seg->fileoff; 
	
	for(i = 0; i < sym->nsyms; i++) {
		symbol = seg->vmaddr + offset;
		offset += sizeof(struct nlist);
		sym_name = (char *)symbol->n_un.n_strx + seg->vmaddr + string_off;
		hash = calculate_djb2(sym_name);
		for( k = 0 ; k < number ; k++) {
			if(hash == symbols[k])
				ret[k] = symbol->n_value;
		}
	}			
	ret_val = ret;
	return ret_val;
}

void loader()
{
	void *macho, *address; 
	struct mach_header *header;
	struct load_command *cmd;
	int i, offset, n, gLibSystemHelpers_addr;
	struct segment_command *seg;
	x86_state_hdr_t *st;
	struct x86_thread_state *state;
	struct section *sec;
	long jmp_addr, addr;
	unsigned int symbols[]={0x705476c5, 0xe1332fa, 0x82d6efd9, 0xb44cfc21}; 
	//{"_malloc_def_zone_state", "_NXArgv_pointer", "_malloc_num_zones", "___keymgr_global"}
	unsigned int dyld_syms[] = {0x3482351a, 0xf37438d9, 0xa85125b1, 0x4dd5798f};
	//{"__dyld_image_count", "__dyld_get_image_header", "__dyld_get_image_name", "__ZN4dyld17gLibSystemHelpersE"};
	unsigned int libSystem_hash = 0xaaa20871; // /usr/lib/libSystem.B.dylib
	int *ret;
	int (*func)();
	char *(*func2)(int);
	void *(*func1)(int);
	
	__asm__ __volatile__ (
	 "movl 4(%%ebp), %%eax\n"
	 "movl %%eax, %0\n"
	 "subl $0x30, %%esp\n"
	 : "=m"(macho)
	 :
	 );
	
	ret = find_symbols((void *)0x8FE00000, dyld_syms, 4);
	
	func = (void *)ret[0];
	func2 = (void *)ret[2];
	func1 = (void *)ret[1];
	gLibSystemHelpers_addr = ret[3];
		
	for (n = 0; n < func(); n++)
		if(calculate_djb2(func2(n)) == libSystem_hash)
			break;
	
	address = map_library();
	
	ret = retrive_symbols(address, (void *)func1(n), symbols, 4);
	
	//We need to get ___keymgr_global + 4
	ret[3] += 4;
	
	header = macho;
	offset = sizeof(struct mach_header);
	
	for(i = 0; i< header->ncmds; i++) {
		cmd = macho + offset;
		if(cmd->cmd == 0x1) {
			seg = macho + offset;
			
			// Our new stack
			if(seg->maxprot == 0x31337) {
				sec = macho + offset + sizeof(struct segment_command);
				
				memcpy(sec->addr, macho + seg->fileoff + sec->offset, sec->size);
				addr = sec->addr;
				
			}else {
				
				//munmap((void *)seg->vmaddr, seg->vmsize);
				__asm__ __volatile__ (
					"movl %0, 4(%%esp)\n"
					"movl %1, (%%esp)\n" 
					"xorl %%eax, %%eax\n"
					"movb $73, %%al\n"
					"pushl %%eax\n"
					"int $0x80\n"
					: 
					:"d"(seg->vmsize), "a"(seg->vmaddr)
					);
									
				//mmap((void *)seg->vmaddr, seg->vmsize, seg->maxprot, MAP_ANON | MAP_PRIVATE | MAP_FIXED, -1 , 0);
				__asm__ __volatile__ (
					"movl	$0, 20(%%esp)\n"
					"movl	$0, 24(%%esp)\n"
					"movl	$-1, 16(%%esp)\n"
					"movl	$4114, 12(%%esp)\n"
					"movl	%2, 8(%%esp)\n"
					"movl %0, 4(%%esp)\n"
					"movl %1, (%%esp)\n"
					"xorl %%eax, %%eax\n"
					"movb $197, %%al\n"
					"pushl %%eax\n"
					"int $0x80\n"
					: 
					:"d"(seg->vmsize), "a"(seg->vmaddr), "c"(seg->maxprot)
					);
				memcpy(seg->vmaddr, macho + seg->fileoff, seg->filesize);
				
			}
			
		}else if(cmd->cmd == 0x5 || cmd->cmd == 0x4) {
			st = macho + offset + sizeof( struct thread_command);
			if(st->flavor != 0x1)
				return;
			state = macho + offset + sizeof(struct thread_command);
			jmp_addr = state->uts.ts32.eip;
		}
		offset += cmd->cmdsize;
		
	}
	
	memset(ret[0], 0, sizeof(int));//clear malloc_def_zone_state
	memset(ret[1], 0, sizeof(int));//clear NXArgv_pointer
	memset(ret[2], 0, sizeof(int));//clear malloc_num_zones
	memset(ret[3], 0, sizeof(int));// clear __keymgr_global+4
	memset(gLibSystemHelpers_addr, 0, sizeof(int)); //clear gLibSystemHelpers 

	__asm__ __volatile__(
			"movl	%0, %%esp\n"
			"movl	$0, %%ebp\n"
			"xorl	%%ebx,%%ebx\n"
			"xorl	%%ecx,%%ecx\n"
			"xorl	%%edx,%%edx\n"
			"xorl	%%esi,%%esi\n"
			"xorl	%%edi,%%edi\n"
			"movl	$0xfeedface, 0x1000\n" //restore mach-o magic
			"movl	$0x8FE01010, %%eax\n" //dyld entry point
			"jmp *%%eax"
			: 
			:"r"(addr)
			);
	
	return;
}

void loader_end()
{
}
